/*	$NetBSD$	*/
/*-
 * Copyright (c) 2010 The NetBSD Foundation, Inc.
 * All rights reserved.
 *
 * This code is derived from software contributed to The NetBSD Foundation
 * by Joerg Sonnenberger.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
 * COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#if HAVE_NBTOOL_CONFIG_H
#include "nbtool_config.h"
#endif

#include <sys/cdefs.h>
__RCSID("$NetBSD$");

#include <sys/endian.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <cdb.h>
#include <errno.h>
#include <fcntl.h>
#include <inttypes.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

struct cdb {
	int fd;
	uint8_t *mmap_base;
	size_t mmap_size;

	uint32_t data_size;
	uint32_t entries;
	uint32_t entries_index;
	uint32_t seed;

	uint8_t offset_size;
	uint8_t index_size;

	uint8_t *hash_base;
	uint8_t *offset_base;
	uint8_t *data_base;
};

struct cdb *
cdb_open(const char *path, int flags)
{
	uint8_t buf[24];
	int fd;
	struct cdb *cdb;
	struct stat sb;

	if ((fd = open(path, O_RDONLY)) == -1)
		return NULL;

	errno = EINVAL;
	if (fstat(fd, &sb) == -1 ||
	    read(fd, buf, sizeof(buf)) != sizeof(buf) ||
	    memcmp(buf, "NBCDB\n\0\001", 8) ||
	    (cdb = malloc(sizeof(*cdb))) == NULL) {
		close(fd);
		return NULL;
	}

	cdb->data_size = le32dec(buf + 8);
	cdb->entries = le32dec(buf + 12);
	cdb->entries_index = le32dec(buf + 16);
	cdb->seed = le32dec(buf + 20);

	if (cdb->data_size < 0x100)
		cdb->offset_size = 1;
	else if (cdb->data_size < 0x10000)
		cdb->offset_size = 2;
	else if (cdb->data_size < 0x1000000)
		cdb->offset_size = 3;
	else
		cdb->offset_size = 4;

	if (cdb->entries_index < 0x100)
		cdb->index_size = 1;
	else if (cdb->entries_index < 0x10000)
		cdb->index_size = 2;
	else if (cdb->entries_index < 0x1000000)
		cdb->index_size = 3;
	else
		cdb->index_size = 4;

	cdb->mmap_size = sb.st_size;
	cdb->mmap_base = mmap(NULL, cdb->mmap_size, PROT_READ, MAP_FILE, fd, 0);
	if (cdb->mmap_base == MAP_FAILED) {
		free(cdb);
		close(fd);
		return NULL;
	}

	cdb->hash_base = cdb->mmap_base + 24;
	cdb->offset_base = cdb->hash_base + cdb->entries_index * cdb->index_size;
	cdb->data_base = cdb->offset_base + (cdb->entries + 1) * cdb->offset_size;

	if (cdb->hash_base < cdb->mmap_base ||
	    cdb->offset_base < cdb->mmap_base ||
	    cdb->data_base < cdb->mmap_base ||
	    cdb->data_base + cdb->data_size < cdb->mmap_base ||
	    cdb->data_base + cdb->data_size >
	    cdb->mmap_base + cdb->mmap_size) {
		cdb_close(cdb);
		return NULL;
	}

	return cdb;
}

static uint32_t
get_uintX(const uint8_t *addr, uint32_t idx, int size)
{
	uint32_t val;

	addr += idx * size;

	val = 0;
	switch (size) {
	case 4:
		val |= addr[3] << 24;
		/* FALLTHROUGH */
	case 3:
		val |= addr[2] << 16;
		/* FALLTHROUGH */
	case 2:
		val |= addr[1] << 8;
		/* FALLTHROUGH */
	case 1:
		val |= addr[0];
	}
	return val;
}

int
cdb_get(struct cdb *cdb, uint32_t idx, const void **data, size_t *data_len)
{
	uint32_t start, end;

	if (idx >= cdb->entries) {
		errno = EINVAL;
		return -1;
	}

	start = get_uintX(cdb->offset_base, idx, cdb->offset_size);
	end = get_uintX(cdb->offset_base, idx + 1, cdb->offset_size);

	if (start > end)
		return -1;

	if (end > cdb->data_size)
		return -1;

	*data = cdb->data_base + start;
	*data_len = end - start;

	return 0;
}



int
cdb_find(struct cdb *cdb, const void *key, size_t key_len,
    const void **data, size_t *data_len)
{
	uint32_t hashes[3], idx;

	if (cdb->entries_index == 0)
		return -1;

	mi_vector_hash(key, key_len, cdb->seed, hashes);

	hashes[0] %= cdb->entries_index;
	hashes[1] %= cdb->entries_index;
	hashes[2] %= cdb->entries_index;

	idx = get_uintX(cdb->hash_base, hashes[0], cdb->index_size);
	idx += get_uintX(cdb->hash_base, hashes[1], cdb->index_size);
	idx += get_uintX(cdb->hash_base, hashes[2], cdb->index_size);

	return cdb_get(cdb, idx % cdb->entries, data, data_len);
}

void
cdb_close(struct cdb *cdb)
{
	munmap(cdb->mmap_base, cdb->mmap_size);
	close(cdb->fd);
	free(cdb);
}
